from .cloud.project import AugerProjectApi
from .cloud.org import AugerOrganizationApi


class Project(AugerProjectApi):
    """Auger Cloud Projects(s) management"""

    def __init__(self, ctx, project_name=None):
        org = AugerOrganizationApi(ctx, ctx.credentials.organization)
        super(Project, self).__init__(ctx, org, project_name)



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerProjectFileApi(AugerBaseApi):
    """Auger Project File Api."""

    def __init__(self, ctx, project_api=None,
        project_file_name=None, project_file_id=None):
        super(AugerProjectFileApi, self).__init__(
            ctx, project_api, project_file_name, project_file_id)
        assert project_api is not None, 'Project must be set for Project File'
        self._set_api_request_path('AugerProjectFileApi')

    def create(self, file_url, file_name=None):
        return self._call_create({
            'name': self.object_name,
            'project_id': self.parent_api.object_id,
            'file_name': file_name, 'url': file_url}, ['processing'])

    def delete(self):
        self.rest_api.call(
            'delete_%s' % self.api_request_path, {'id': self.oid})



class AugerConfig(object):
    """Modify configuration options in auger.yaml."""

    def __init__(self, ctx):
        super(AugerConfig, self).__init__()
        self.ctx = ctx

    def _with_auger_config(decorated):
        def wrapper(self, *args, **kwargs) :
            res = decorated(self, *args, **kwargs)
            self.ctx.config.write('auger')

            return res
            
        return wrapper

    @_with_auger_config
    def set_data_set(self, name, source=None, validation=False):
        #config.set('auger', 'dataset', data_set_name)

        if validation:
            self.ctx.config.set('auger', 'validation_dataset', name)
        else:
            self.ctx.config.set('auger', 'dataset', name)

        if source:
            self.ctx.config.set('config', 'source', source)

        if self.ctx.config.ismultipart():
            self.ctx.config.write('config')

        return self
            
    @_with_auger_config
    def set_experiment(self, experiment_name, experiment_session_id):
        self.ctx.config.set('auger', 'experiment/name', experiment_name)
        self.ctx.config.set('auger', 'experiment/experiment_session_id',
            experiment_session_id)

    @_with_auger_config
    def set_project(self, project_name):
        self.ctx.config.set('config', 'name', project_name)
        if (self.ctx.config.ismultipart()):
            self.ctx.config.write('config')
        return self

    # @_with_auger_config
    # def config(self, config, *args, **kwargs):
    #     config.set('auger', 'experiment/name',
    #         kwargs.get('experiment_name', ''))
    #     config.set('auger', 'dataset',
    #         kwargs.get('data_set_name', ''))
    #     config.set('auger', 'experiment/experiment_session_id',
    #         kwargs.get('experiment_session_id', ''))

        # model_type = kwargs.get('model_type', None)
        # if model_type:
        #     yaml['experiment']['metric'] = \
        #         'f1_macro' if model_type == 'classification' else 'r2'

        # self.ctx.config.set('config', 'name', kwargs.get('project_name', ''))

        # self.ctx.config.set('config', 'source', kwargs.get('source', ''))
        # self.ctx.config.set('auger', 'dataset', kwargs.get('data_set_name', ''))
        # self.ctx.config.set('config', 'target', kwargs.get('target', ''))

        # self.ctx.config.set('auger', 'experiment/name',
        #     kwargs.get('experiment_name', ''))
        # model_type = kwargs.get('model_type', None)
        # if model_type:
        #     self.ctx.config.set('auger', 'experiment/metric',
        #         'f1_macro' if model_type == 'classification' else 'r2')
        # self.ctx.config.set('config', 'model_type', model_type or '')
        # if (self.ctx.config.ismultipart()):
        #     self.ctx.config.write('config')



import os
import datetime
from dateutil.tz import tzutc
import mimetypes
import logging


class S3FSClient:

    def _get_relative_path(self, path):
        try:
            from urllib.parse import urlparse
        except ImportError:
            from urlparse import urlparse

        relPath = path
        uri = urlparse(path)
        self.s3BucketName = uri.netloc

        self.client = self._create_botos3_client()
        relPath = uri.path[1:]

        relPath = relPath.replace("//", "/")
        if relPath.endswith("/"):
            relPath = relPath[:-1]

        return relPath

    @staticmethod
    def _create_botos3_client():
        import boto3

        endpoint_url = os.environ.get('S3_ENDPOINT_URL')

        client = None
        if endpoint_url:
            client = boto3.client(
                's3',
                endpoint_url=endpoint_url,
                config=boto3.session.Config(signature_version='s3v4')
            )
        else:
            client = boto3.client(
                's3',
                config=boto3.session.Config(signature_version='s3v4')
            )

        return client

    def get_smart_open_transport_params(self):
        transport_params = None
        endpoint_url = os.environ.get('S3_ENDPOINT_URL')

        if endpoint_url:
            transport_params = {
                'resource_kwargs': {
                    'endpoint_url': endpoint_url,
                }
            }

        return transport_params

    def wait_for_path(self, path):
        path = self._get_relative_path(path)
        if 'object_exists' in self.client.waiter_names:
            waiter = self.client.get_waiter('object_exists')
            waiter.config.delay = 2
            waiter.config.max_attempts = 50

            logging.info("S3FSClient: waiting for max %s sec for object: %s" % (
                waiter.config.delay*waiter.config.max_attempts, path))
            waiter.wait(Bucket=self.s3BucketName, Key=path)

    def s3fs_open(self, path, mode):
        from s3fs.core import S3FileSystem

        endpoint_url = os.environ.get('S3_ENDPOINT_URL')
        client_kwargs = {}
        if endpoint_url:
            client_kwargs = {'endpoint_url': endpoint_url}

        if 'r' in mode:
            self.wait_for_path(path)

        s3 = S3FileSystem(anon=False, default_fill_cache=False,
                          client_kwargs=client_kwargs)
        return s3.open(path, mode=mode)

    def list_folder(self, path, wild=False, removeFolderName=False, meta_info=False):
        import fnmatch
        from urllib.parse import unquote

        path = self._get_relative_path(path)
        path_filter = None
        if path and wild:
            parts = path.split('/')
            path = '/'.join(parts[:-1])
            path_filter = parts[-1]

        if path:
            path = path + "/" if not path.endswith("/") else path

        #print("s3BucketName: %s;Path:%s; path_filter:%s"%(self.s3BucketName, path, path_filter))

        ContinuationToken = None
        result = []
        while True:
            if ContinuationToken:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/', ContinuationToken=ContinuationToken)
            else:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/')

            ContinuationToken = listFiles.get('NextContinuationToken', None)
            files = None
            folders = None
            if listFiles is not None:
                files = listFiles.get('Contents')
                folders = listFiles.get('CommonPrefixes')

            if files is not None:
                for file in files:
                    if file.get('Key') != path:
                        file_path = unquote(file.get('Key')[len(path):])
                        if meta_info:
                            result.append({'path': file_path,
                                           'last_modified': self._get_seconds_from_epoch(file.get('LastModified')), 'size': file.get('Size', 0)})
                        else:
                            result.append(file_path)

            if folders is not None:
                for folder in folders:
                    if folder.get('Prefix') != path:
                        folder_path = unquote(folder.get('Prefix')[len(path):])
                        if meta_info:
                            result.append({'path': folder_path,
                                           'last_modified': self._get_seconds_from_epoch(folder.get('LastModified')), 'size': folder.get('Size', 0)})
                        else:
                            result.append(folder_path)

            if not ContinuationToken:
                break

        if path_filter:
            if meta_info:
                result = [item for item in result if fnmatch.fnmatch(
                    item['path'], path_filter)]
            else:
                result = fnmatch.filter(result, path_filter)

        return result

    def create_folder(self, path):
        path = self._get_relative_path(path)
        path = path + "/" if not path.endswith("/") else path

        # If no path we shouldn't create it
        if path != "/":
            self.client.put_object(Bucket=self.s3BucketName, Key=path)

    def create_parent_folder(self, path):
        parent = os.path.dirname(path)
        self.create_folder(parent)

    def remove_folder(self, path, remove_self=True):
        path = self._get_relative_path(path)
        self._s3_removeFolder(path, remove_self)

    def remove_file(self, path, wild=False):
        if wild:
            files = self.list_folder(path, wild)
            path = self._get_relative_path(path)
            for file in files:
                path_to_remove = os.path.join(os.path.dirname(path), file)
                # print(path_to_remove)
                self.client.delete_object(
                    Bucket=self.s3BucketName, Key=path_to_remove)

            #raise Exception("S3FSClient::remove_file wild is not implemented:%s"%path)
        else:
            path = self._get_relative_path(path)
            self.client.delete_object(Bucket=self.s3BucketName, Key=path)

    def _s3_removeFolder(self, path, remove_self=True):
        path = path + "/" if not path.endswith("/") else path

        # print("Remove:%s"%path)
        ContinuationToken = None
        while True:
            if ContinuationToken:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/', ContinuationToken=ContinuationToken)
            else:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/')

            # print(listFiles)
            ContinuationToken = listFiles.get('NextContinuationToken', None)
            files = None
            folders = None
            if listFiles is not None:
                files = listFiles.get('Contents')
                folders = listFiles.get('CommonPrefixes')

            if files is not None:
                for file in files:
                    #print("Delete path:%s"%file.get('Key'))
                    self.client.delete_object(
                        Bucket=self.s3BucketName, Key=file.get('Key'))

            if folders is not None:
                for folder in folders:
                    self._s3_removeFolder(folder.get(
                        'Prefix'), remove_self=False)

            if not ContinuationToken:
                break

        if remove_self:
            self.client.delete_object(Bucket=self.s3BucketName, Key=path[:-1])

    def _s3_moveFile(self, oldName, newName):
        self.client.copy_object(Bucket=self.s3BucketName, CopySource={
                                'Bucket': self.s3BucketName, 'Key': oldName}, Key=newName)
        self.client.delete_object(Bucket=self.s3BucketName, Key=oldName)

    def is_file_exists(self, path):
        path = self._get_relative_path(path)
        exists = False
        try:
            res = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            exists = True
        except Exception as e:
            pass

        return exists

    @staticmethod
    def _get_seconds_from_epoch(date_data):
        res = 0
        try:
            if date_data:
                epoch = datetime.datetime(1970, 1, 1, tzinfo=tzutc())
                res = (date_data - epoch).total_seconds()
        except Exception as e:
            #logging.exception("_get_seconds_from_epoch failed.")
            pass

        return res

    def get_mtime(self, path):
        path = self._get_relative_path(path)
        res = None
        try:
            obj = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            res = self._get_seconds_from_epoch(obj.get('LastModified'))
        except Exception as e:
            #logging.exception("getMTime failed.")
            pass

        return res

    def get_file_size(self, path):
        path = self._get_relative_path(path)
        res = 0
        try:
            obj = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            res = obj.get('ContentLength')
        except Exception as e:
            #logging.exception("get_file_size failed.")
            pass

        return res

    def is_folder_exists(self, path):
        path = self._get_relative_path(path)
        listFiles = self.client.list_objects(
            Bucket=self.s3BucketName, Prefix=path + "/", Delimiter='/')
        if listFiles is not None:
            content = listFiles.get('Contents')
            if content is None:
                listFiles = listFiles.get('CommonPrefixes')
            else:
                listFiles = content

        # print(listFiles)
        return listFiles is not None

    def read_text_file(self, path):
        from .LocalFSClient import LocalFSClient

        with LocalFSClient().save_atomic(path) as local_tmp_path:
            self.download_file(path, local_tmp_path)
            return LocalFSClient().read_text_file(local_tmp_path)

        # obj = self.client.get_object(Bucket=self.s3BucketName, Key=path)
        # return obj['Body'].read()

    def write_text_file(self, path, data, atomic=False):
        path = self._get_relative_path(path)
        mimetype, _ = mimetypes.guess_type(path)
        if mimetype is None:
            self.client.put_object(
                Body=data, Bucket=self.s3BucketName, Key=path)
        else:
            self.client.put_object(
                Body=data, Bucket=self.s3BucketName, Key=path, ContentType=mimetype)

    def copy_file_remote(self, path_src, path_dst):
        path = self._get_relative_path(path_src)
        copy_source = {
            'Bucket': self.s3BucketName,
            'Key': path
        }

        path = self._get_relative_path(path_dst)
        self.client.copy(copy_source, self.s3BucketName, path)

    def copy_file(self, path_src, path_dst):
        if path_src.startswith("s3") and path_dst.startswith("s3"):
            self.copy_file_remote(path_src, path_dst)
        elif path_dst.startswith("s3"):
            path = self._get_relative_path(path_dst)
            self._s3_upload_file(path_src, path)
        elif path_src.startswith("s3"):
            self.download_file(path_src, path_dst)

    def _s3_upload_file(self, path_local, path_s3):
        import boto3

        s3_config = boto3.s3.transfer.TransferConfig(use_threads=False)

        mimetype, _ = mimetypes.guess_type(path_local)
        if mimetype is None:
            self.client.upload_file(
                path_local, Bucket=self.s3BucketName, Key=path_s3, Config=s3_config)
        else:
            self.client.upload_file(path_local, Bucket=self.s3BucketName, Key=path_s3, Config=s3_config,
                                    ExtraArgs={"ContentType": mimetype}
                                    )

    def copy_files(self, path_src, path_dst):
        files = self.list_folder(path_src, wild=True)
        for file in files:
            self.copy_file(os.path.join(os.path.dirname(
                path_src), file), os.path.join(path_dst, file))

    def copy_folder(self, path_src, path_dst):
        from .FSClient import FSClient

        files = FSClient().list_folder(path_src)
        for file in files:
            full_src = os.path.join(path_src, file)
            if FSClient().is_file_exists(full_src):
                self.copy_file(full_src, os.path.join(path_dst, file))
            else:
                self.copy_folder(full_src, os.path.join(path_dst, file))

    def download_file(self, path, local_path):
        from .LocalFSClient import LocalFSClient
        from .FSClient import FSClient
        import boto3
        try:
            from urllib.parse import urlparse
        except ImportError:
            from urlparse import urlparse

        if path.startswith("http:") or path.startswith("https:"):
            s3_path = self._get_relative_path(local_path)

            uri = urlparse(path)

            with LocalFSClient().save_atomic(uri.path) as temp_file:
                LocalFSClient().downloadFile(path, temp_file)
                self._s3_upload_file(temp_file, s3_path)

                #FSClient().waitForFile(local_path, wait_for_file=True, num_tries=3000, interval_sec=20)
            # with FSClient().open(path, "rb", encoding=None) as fd:
            #     self.client.upload_fileobj(fd, Bucket=self.s3BucketName, Key=s3_path)
        else:
            path = self._get_relative_path(path)
            LocalFSClient().create_parent_folder(local_path)
            self.client.download_file(
                Bucket=self.s3BucketName, Key=path, Filename=local_path)

    def download_folder(self, path, local_path):
        files = self.list_folder(path)
        for file in files:
            if file.endswith('/'):
                self.download_folder(os.path.join(
                    path, file), os.path.join(local_path, file))
            else:
                self.copy_file(os.path.join(path, file),
                               os.path.join(local_path, file))

    def move_file(self, path_src, path_dst):
        self.copy_file(path_src, path_dst)

        if path_src.startswith("s3"):
            self.remove_file(path_src)
        else:
            from .LocalFSClient import LocalFSClient
            LocalFSClient().remove_file(path_src)



import os
import shutil
import subprocess
from zipfile import ZipFile

from .deploy import ModelDeploy
from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..exceptions import AugerException
from a2ml.api.utils.dataframe import DataFrame


class ModelPredict():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelPredict, self).__init__()
        self.ctx = ctx

    def execute(self, filename, model_id, threshold=None, locally=False):
        self.ctx.log('Predicting on data in %s' % filename)
        filename = os.path.abspath(filename)

        if locally:
            predicted = self._predict_locally(filename, model_id, threshold)
        else:
            predicted = self._predict_on_cloud(filename, model_id, threshold)

        self.ctx.log('Predictions stored in %s' % predicted)

        return predicted

    def _predict_on_cloud(self, filename, model_id, threshold=None):
        target = self.ctx.config.get('target', None)
        records, features = DataFrame.load_records(filename, target)

        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        predictions = pipeline_api.predict(
            records, features, threshold)

        predicted = os.path.splitext(filename)[0] + "_predicted.csv"
        DataFrame.save(predicted, predictions)

        return predicted

    def _predict_locally(self, filename, model_id, threshold):
        model_deploy = ModelDeploy(self.ctx, None)
        is_model_loaded, model_path, model_name = \
            model_deploy.verify_local_model(model_id)

        if not is_model_loaded:
            raise AugerException('Model isn\'t loaded locally. '
                'Please use a2ml deploy command to download model.')

        model_path, model_existed = self._extract_model(model_name)

        try:
            predicted = \
                self._docker_run_predict(filename, threshold, model_path)
        finally:
            # clean up unzipped model
            # if it wasn't unzipped before
            if not model_existed:
                shutil.rmtree(model_path, ignore_errors=True)

        return predicted

    def _extract_model(self, model_name):
        model_path = os.path.splitext(model_name)[0]
        model_existed = os.path.exists(model_path)

        if not model_existed:
            with ZipFile(model_name, 'r') as zip_file:
                zip_file.extractall(model_path)

        return model_path, model_existed

    def _docker_run_predict(self, filename, threshold, model_path):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')
        result_file = os.path.basename(filename)
        data_path = os.path.dirname(filename)

        call_args = "--path_to_predict=./model_data/%s %s" % \
            (result_file, "--threshold=%s" % str(threshold) if threshold else '')

        command = (r"docker run "
            "-v {model_path}:/var/src/auger-ml-worker/exported_model "
            "-v {data_path}:/var/src/auger-ml-worker/model_data "
            "deeplearninc/auger-ml-worker:{docker_tag} "
            "python ./exported_model/client.py {call_args}").format(
                model_path=model_path, data_path=data_path,
                docker_tag=docker_tag, call_args=call_args)

        try:
            self.ctx.log(
                'Running model in deeplearninc/'
                'auger-ml-worker:%s' % docker_tag)
            getattr(subprocess,
                'check_call' if self.ctx.debug else 'check_output')(
                command, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Error running Docker container...')

        return os.path.join(data_path,
            os.path.splitext(result_file)[0] + "_predicted.csv")



from .base import AugerBaseApi


class AugerActualApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api, prediction_id=None):
        super(AugerActualApi, self).__init__(
            ctx, pipeline_api, prediction_id)

    def create(self, records):
        return self._call_create(params={
            'pipeline_id': self.parent_api.object_id,
            'actuals': records}, has_return_object=False)



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('review', short_help='Review specified model info.')
@pass_context
def cmdl(ctx):
    """Review specified model info."""
    ctx.setup_logger(format='')
    A2ML(ctx).review()



import os
import csv
from ..cloud.pipeline import AugerPipelineApi


class ModelActual():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelActual, self).__init__()
        self.ctx = ctx

    def execute(self, filename, model_id):
        self.ctx.log('Sending actuals on data in %s' % filename)
        filename = os.path.abspath(filename)
        self._actual_to_cloud(filename, model_id)
        return True

    def _actual_to_cloud(self, filename, model_id):
        with open(filename) as f:
          a = [{k: v for k, v in row.items()}
              for row in csv.DictReader(f, skipinitialspace=True)]

        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        pipeline_api.actual(a)



import os

class Config:
    def __init__(self):
        self.celery_broker_url = os.environ.get('CELERY_BROKER_URL', 'redis://localhost:6379/0')
        self.debug = os.environ.get('DEBUG', 'false').lower() == 'true'
        self.notificator_redis_host = os.environ.get('NOTIFICATOR_REDIS_HOST', 'localhost')
        self.notificator_redis_port = int(os.environ.get('NOTIFICATOR_REDIS_PORT', 6379))



from .utils import interceptor, EXPERIMENT, PROJECT_FILE
from a2ml.api.auger.impl.cloud.experiment import AugerExperimentApi


class TestConfigs(object):

    def test_experiment_settings(self, project, ctx_api, monkeypatch):
        config = ctx_api.config
        config.set('config','target', 'species')
        config.set('config','experiment/cross_validation_folds', 55)
        config.set('config','experiment/max_total_time', 606)
        config.set('config','experiment/max_eval_time', 55)
        config.set('config','experiment/max_n_trials', 101)
        config.set('config','experiment/use_ensemble', False)

        PAYLOAD = {
            'get_experiment': EXPERIMENT,
            'get_project_file': PROJECT_FILE
        }
        interceptor(PAYLOAD, monkeypatch)
        config, model_type = AugerExperimentApi(
            ctx_api, 'project-api', 'iris.csv-experiment', '1234').\
            get_experiment_settings()

        assert config['evaluation_options']['crossValidationFolds'] == 55
        assert config['evaluation_options']['max_total_time_mins'] == 606
        assert config['evaluation_options']['max_eval_time_mins'] == 55
        assert config['evaluation_options']['max_n_trials'] == 101
        assert config['evaluation_options']['use_ensemble'] == False
        # dataset
        assert config['evaluation_options']['targetFeature'] == 'species'
        assert config['evaluation_options']['featureColumns'] == \
            ['sepal_length', 'sepal_width', 'petal_length', 'petal_width']
        assert config['evaluation_options']['categoricalFeatures'] == \
            ['species']
        assert config['evaluation_options']['timeSeriesFeatures'] == []
        assert config['evaluation_options']['binaryClassification'] == False
        assert config['evaluation_options']['labelEncodingFeatures'] == []
        assert config['evaluation_options']['classification'] == True
        assert config['evaluation_options']['scoring'] == 'f1_macro'

    def test_exclude_setting(self, project, ctx_api, monkeypatch):
        config = ctx_api.config
        config.set('config','target', 'species')
        config.set('config','exclude',['sepal_length'])

        PAYLOAD = {
            'get_experiment': EXPERIMENT,
            'get_project_file': PROJECT_FILE
        }
        interceptor(PAYLOAD, monkeypatch)
        config, model_type = AugerExperimentApi(
            ctx_api, 'project-api', 'iris.csv-experiment', '1234').\
            get_experiment_settings()

        assert config['evaluation_options']['targetFeature'] == 'species'
        assert config['evaluation_options']['featureColumns'] == \
            ['sepal_width', 'petal_length', 'petal_width']
        assert config['evaluation_options']['categoricalFeatures'] == \
            ['species']

    def test_model_type_setting(self, project, ctx_api, monkeypatch):
        ctx_api.config.set('config','target', 'species')
        ctx_api.config.set('config','model_type','regression')
        ctx_api.config.set('auger','experiment/metric', None)

        PAYLOAD = {
            'get_experiment': EXPERIMENT,
            'get_project_file': PROJECT_FILE
        }
        interceptor(PAYLOAD, monkeypatch)
        config, model_type = AugerExperimentApi(
            ctx_api, 'project-api', 'iris.csv-experiment', '1234').\
            get_experiment_settings()

        assert config['evaluation_options']['timeSeriesFeatures'] == []
        assert config['evaluation_options']['binaryClassification'] == False
        assert config['evaluation_options']['labelEncodingFeatures'] == []
        assert config['evaluation_options']['classification'] == False
        assert config['evaluation_options']['scoring'] == 'r2'



import os

from a2ml.api.utils import fsclient


class BaseCredentials(object):
    """Manage credentials on user computer."""
    def __init__(self, ctx, provider):
        super(BaseCredentials, self).__init__()
        self.ctx = ctx
        self.provider = provider
        self.creds_path = self._path_to_credentials()
        self.creds_file = os.path.join(self.creds_path, '%s.json'%self.provider)

    def _path_to_credentials(self):
        if self.ctx.config.get('path_to_credentials'):
            creds_path = os.path.abspath(self.ctx.config.get('path_to_credentials'))
        elif os.environ.get('%s_CREDENTIALS_PATH'%self.provider.upper()):
            creds_path = os.environ.get('%s_CREDENTIALS_PATH'%self.provider.upper())
        else:
            cur_path = os.getcwd()
            if self.ctx.config.path:
                cur_path = self.ctx.config.path

            if fsclient.is_file_exists(os.path.join(cur_path, "%s.json"%self.provider)):
                creds_path = cur_path
            else:
                creds_path = os.path.abspath('%s/.a2ml' % os.environ.get('HOME', os.getcwd()))

        return creds_path

    def _credentials_file_exist(self):
        return os.path.exists(self.creds_file)

    def _ensure_credentials_file(self):
        if not os.path.exists(self.creds_path):
            os.makedirs(self.creds_path)

        if not os.path.exists(self.creds_file):
            with open(self.creds_file, 'w') as f:
                f.write('{}')



from functools import wraps

def show_result(decorated):
    @wraps(decorated)
    def wrapper(self, *args, **kwargs):
        showresult = self.ctx.config.get('showresult', False)
        result = decorated(self, *args, **kwargs)
        if showresult:
            self.ctx.log(result)
        # else:
        # 	for provider, res in result.items():
        # 		if "result" in res and not res["result"]:
        # 			self.ctx.log("Provider '%s' return error: %s"%(provider, res.get("data")))		

        return result
    return wrapper



from .impl.cloud.rest_api import RestApi
from .impl.decorators import error_handler, authenticated, with_dataset
from .impl.experiment import Experiment
from .impl.exceptions import AugerException
from a2ml.api.utils.formatter import print_table
from .credentials import Credentials
from .config import AugerConfig

class AugerExperiment(object):

    def __init__(self, ctx):
        self.ctx = ctx
        self.ctx.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.ctx.credentials.api_url, self.ctx.credentials.token)

    @error_handler
    @authenticated
    @with_dataset
    def list(self, dataset):
        count = 0
        for exp in iter(Experiment(self.ctx, dataset).list()):
            self.ctx.log(exp.get('name'))
            count += 1
        self.ctx.log('%s Experiment(s) listed' % str(count))
        return {'experiments': Experiment(self.ctx, dataset).list()}

    @error_handler
    @authenticated
    @with_dataset
    def start(self, dataset):
        experiment_name = \
            self.ctx.config.get('experiment/name', None)
        experiment_name, session_id = \
            Experiment(self.ctx, dataset, experiment_name).start()
        AugerConfig(self.ctx).set_experiment(experiment_name, session_id)
        return {'experiment_name': experiment_name, 'session_id': session_id}

    @error_handler
    @authenticated
    @with_dataset
    def stop(self, dataset, run_id = None):
        name = self.ctx.config.get('experiment/name', None)
        if name is None:
            raise AugerException('Please specify Experiment name...')
        if run_id is None:
            run_id = self.ctx.config.get(
                'experiment/experiment_session_id', None)

        if Experiment(self.ctx, dataset, name).stop(run_id):
            self.ctx.log('Search is stopped...')
        else:
            self.ctx.log('Search is not running. Stop is ignored.')
        return {'stopped': name}

    @error_handler
    @authenticated
    @with_dataset
    def leaderboard(self, dataset, run_id = None):
        name = self.ctx.config.get('experiment/name', None)
        if name is None:
            raise AugerException('Please specify Experiment name...')
        if run_id is None:
            run_id = self.ctx.config.get(
                'experiment/experiment_session_id', None)
        leaderboard, status, run_id = Experiment(
            self.ctx, dataset, name).leaderboard(run_id)
        if leaderboard is None:
            raise AugerException('No leaderboard was found...')
        self.ctx.log('Leaderboard for Run %s' % run_id)
        print_table(self.ctx.log, leaderboard[::-1])
        messages = {
            'preprocess': 'Search is preprocessing data for training...',
            'started': 'Search is in progress...',
            'completed': 'Search is completed.',
            'interrupted': 'Search was interrupted.'
        }
        message = messages.get(status, None)
        if message:
            self.ctx.log(message)
        else:
            self.ctx.log('Search status is %s' % status)
        return {'run_id': run_id, 'leaderboard': leaderboard, 'status': status}

    @error_handler
    @authenticated
    @with_dataset
    def history(self, dataset):
        name = self.ctx.config.get('experiment/name', None)
        if name is None:
            raise AugerException('Please specify Experiment name...')
        for exp_run in iter(Experiment(self.ctx, dataset, name).history()):
            self.ctx.log("run id: {}, start time: {}, status: {}".format(
                exp_run.get('id'),
                exp_run.get('model_settings').get('start_time'),
                exp_run.get('status')))
        return {'history': Experiment(self.ctx, dataset, name).history()}



from .mparts.deploy import ModelDeploy
from .mparts.predict import ModelPredict
from .mparts.actual import ModelActual


class Model(object):
    """Auger Cloud Model(s) management."""

    def __init__(self, ctx, project):
        super(Model, self).__init__()
        self.project = project
        self.ctx = ctx

    def deploy(self, model_id, locally=False):
        return ModelDeploy(self.ctx, self.project).execute(model_id, locally)

    def predict(self, filename, model_id, threshold=None, locally=False):
        if locally:
            self.deploy(model_id, locally)
            
        return ModelPredict(self.ctx).execute(filename, model_id, threshold, locally)

    def actual(self, filename, model_id):
        return ModelActual(self.ctx).execute(filename, model_id)



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerClusterApi(AugerBaseApi):
    """Auger Cluster API."""

    def __init__(self, ctx, project_api, cluster_id=None):
        super(AugerClusterApi, self).__init__(
            ctx, project_api, None, cluster_id)
        assert project_api is not None, 'Project must be set for Cluster'

    def is_running(self):
        if self.object_id is None:
            return False
        return self.properties().get('status') == 'running'

    def create(self):
        params = {
            'project_id': self.parent_api.object_id,
            'organization_id': self.parent_api.parent_api.object_id}
        params.update(self.get_cluster_settings(self.ctx))
        return self._call_create(params,
            ['waiting', 'provisioning', 'bootstrapping'])

    @staticmethod
    def get_cluster_settings(ctx):
        config = ctx.config

        default_stack = "stable"
        if 'staging' in ctx.rest_api.api_url:
            default_stack = 'experimental'

        kubernetes_stack = config.get('cluster/kubernetes_stack', None)
        if kubernetes_stack is None:
            kubernetes_stack = config.get(
                'cluster/stack_version', default_stack)

        settings = {
            "kubernetes_stack": kubernetes_stack,
            "autoterminate_minutes":
                config.get('cluster/autoterminate_minutes', 30)
        }

        docker_image_tag = config.get('cluster/docker_image_tag', None)
        if docker_image_tag is not None:
            settings["docker_image_tag"] = docker_image_tag

        cluster_type = config.get('cluster/type', None)
        if cluster_type is not None:
            if cluster_type not in ['standard', 'high_memory']:
                raise AugerException(
                    'Cluster type \'%s\' is not supported' % cluster_type)
            settings.update({
                "worker_type_id": 1 if cluster_type == 'standard' else 2,
                "workers_count": config.get('cluster/max_nodes', 2),
            })
        else: # single tenant settings
            worker_nodes_count = config.get('cluster/worker_count', 2)
            worker_nodes_count = config.get(
                'cluster/worker_nodes_count', worker_nodes_count)

            settings.update({
                "worker_nodes_count": worker_nodes_count,
                "instance_type": config.get('cluster/instance_type', 'c5.large')
            })
            workers_per_node_count = config.get(
                'cluster/workers_per_node_count', None)
            if workers_per_node_count is not None:
                settings["workers_per_node_count"] = workers_per_node_count

        return settings



import os
import logging
import ruamel.yaml

from a2ml.api.utils.config_yaml import ConfigYaml
from a2ml.api.utils import fsclient

log = logging.getLogger("a2ml")

'''Config to serialize pass to server side deserialize, then pass back and save on CLI side'''
class SerializableConfigYaml(ConfigYaml):
    def __getstate__(self):
        return {
            'filename': self.filename,
            'yaml': ruamel.yaml.dump(self.yaml, Dumper=ruamel.yaml.RoundTripDumper)
        }

    def __setstate__(self, state):
        self.filename = state['filename']
        self.yaml = ruamel.yaml.load(state['yaml'], Loader=ruamel.yaml.RoundTripLoader)

    def write(self, filename=None, client_side=True):
        if client_side:
            super().write(filename)


class ConfigParts(object):
    def __init__(self):
        self.parts = {}
        self.part_names = ['config', 'auger', 'azure', 'google']
        self.is_loaded = False

    def load(self, path=None):
        if path is None:
            path = os.getcwd()
        for pname in self.part_names:
            filename = os.path.join(path, '%s.yaml' % pname)

            if not fsclient.is_s3_path(filename):
                filename = os.path.abspath(filename)

            if fsclient.is_file_exists(filename):
                self.parts[pname] = self._load(filename)
        self.is_loaded = True

    def ismultipart(self):
        return(len(self.parts.keys()) > 1)

    def keys(self):
        return self.parts.keys()

    def part(self, name):
        return self.parts[name]

    def _load(self, name):
        part = SerializableConfigYaml()
        if fsclient.is_file_exists(name):
            part.load_from_file(name)
        return part

class Config(object):

    def __init__(self, name='config', path=None):
        super(Config, self).__init__()
        self.runs_on_server = False
        self.name = name
        self.path = path
        self.parts = ConfigParts()
        self.load(path)

    def get(self, path, default=None):
        if len(self.parts.keys()) == 0:
            return default
        if 'config' in self.parts.keys():
            default = self.parts.part('config').get(path, default)
        return self.parts.part(self.name).get(path, default)

    def get_list(self, path, default=None):
        data = self.get(path, default)
        res = None

        if data:
            if isinstance(data, str):
                res = data.split(",")
                res = [item.strip() for item in res]
            else:
                res = list(data)

        return res
               
    def get_path(self):
        path = self.path
        if path is None:
            path = os.getcwd()

        return path
                   
    def set(self, part_name, path, value):
        if (part_name == 'config' and self.ismultipart()):
            self.parts.part('config').set(path, value)
        else:
            self.parts.part(self.name).set(path, value)

    def write(self, part_name):
        if (part_name == 'config' and self.ismultipart()):
            self.parts.part('config').write(client_side=not self.runs_on_server)
        else:
            self.parts.part(self.name).write(client_side=not self.runs_on_server)

    def write_all(self):
        for part_name in self.parts.parts.keys():
            self.write(part_name)

    def ismultipart(self):
        return self.parts.ismultipart()

    def load(self, path=None, reload=False):
        if (not self.parts.is_loaded) or reload:
            self.parts.load(path)
        return self



from .base import AugerBaseApi
from .dataset import AugerDataSetApi
from .experiment_session import AugerExperimentSessionApi
from ..exceptions import AugerException

MODEL_TYPES = ['classification', 'regression', 'timeseries']

class AugerExperimentApi(AugerBaseApi):
    """Auger Experiment Api."""

    def __init__(self, ctx, project_api,
        experiment_name=None, experiment_id=None):
        super(AugerExperimentApi, self).__init__(
            ctx, project_api, experiment_name, experiment_id)
        assert project_api is not None, 'Project must be set for Experiment'
        self._set_api_request_path('AugerExperimentApi')

    def run(self):
        experiment_session_api = \
            AugerExperimentSessionApi(self.ctx, self)
        experimeny_session_properties = \
            experiment_session_api.create()
        experiment_session_api.run()
        return experimeny_session_properties.get('id')

    def create(self, data_set_name):
        assert data_set_name is not None, \
            'DataSet Name is required to create Experiment'

        data_set_api = AugerDataSetApi(
            self.ctx, self.parent_api, data_set_name)
        data_set_properties = data_set_api.properties()

        if not self.object_name:
            self.object_name = self._get_uniq_object_name(
                data_set_name, '-experiment')

        return self._call_create({
            'name': self.object_name,
            'project_id': self.parent_api.object_id,
            'data_path': data_set_properties.get('url')})

    def get_experiment_settings(self):
        config = self.ctx.config

        model_type = config.get('model_type', '')
        if not model_type in MODEL_TYPES:
            raise AugerException('Model type should be %s' % \
                '|'.join(MODEL_TYPES))
        target = config.get('target', '')
        exclude = config.get('exclude', [])

        options = {
            'targetFeature': None,
            'featureColumns': [],
            'categoricalFeatures': [],
            'timeSeriesFeatures': [],
            'binaryClassification': False,
            'labelEncodingFeatures':
                config.get('experiment/label_encoded', []),
            'crossValidationFolds':
                config.get('experiment/cross_validation_folds', 5),
            'max_total_time_mins':
                config.get('experiment/max_total_time', 60),
            'max_eval_time_mins':
                config.get('experiment/max_eval_time', 6),
            'max_n_trials':
                config.get('experiment/max_n_trials', 1000),
            'use_ensemble':
                config.get('experiment/use_ensemble', True),
            'classification':
                True if model_type == 'classification' else False,
            'scoring':
                config.get('experiment/metric',
                    'f1_macro' if model_type == 'classification' else 'r2')
        }

        data_set_id = self.properties()['project_file_id']
        data_set_api = AugerDataSetApi(
            self.ctx, self.parent_api, None, data_set_id)
        data_set_properties = data_set_api.properties()
        stats = data_set_properties['statistics']

        self._fill_data_options(options, stats, target, exclude)

        if options['targetFeature'] is None:
            raise AugerException('Please set target to build model.')

        if model_type is not 'timeseries':
            options['timeSeriesFeatures'] = []
        else:
            time_series = config.get('experiment/time_series', None)
            if time_series:
                options['timeSeriesFeatures'] = [time_series]
            if len(options['timeSeriesFeatures']) != 1:
                raise AugerException('Please select time series feature'
                    ' to build time series model'
                    ' (experiment/time_series option).')

        return {'evaluation_options': options}, model_type

    def _fill_data_options(self, options, stats, target, exclude):
        for item in stats.get('stat_data'):
            column_name = item['column_name']
            if column_name in exclude:
                continue
            if column_name != target:
                options['featureColumns'].append(column_name)
            else:
                options['targetFeature'] = target
            if item['datatype'] == 'categorical':
                options['categoricalFeatures'].append(column_name)
                if column_name == target:
                    options['binaryClassification'] = \
                        True if item['unique_values'] == 2 else False
            if item['datatype'] == 'date':
                options['timeSeriesFeatures'].append(column_name)
                options['datetime_features'].append(column_name)



import os
import csv

from .impl.cloud.pipeline import AugerPipelineApi
from .impl.decorators import error_handler, authenticated, with_project


class AugerActual(object):
    """Actual value for deployed Auger Pipeline."""

    def __init__(self, ctx):
        super(AugerActual, self).__init__(ctx)
        self.ctx.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.ctx.credentials.api_url, self.ctx.credentials.token)
        
    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def actual(self, filename, model_id):
        self.ctx.log('Sending actuals on data in %s' % filename)
        filename = os.path.abspath(filename)
        self._actual_to_cloud(filename, model_id)

    def _actual_to_cloud(self, filename, model_id):
        with open(filename) as f:
          a = [{k: v for k, v in row.items()} 
              for row in csv.DictReader(f, skipinitialspace=True)]

        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
        pipeline_api.actual(a)